#
# Copyright (C) 2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Customized logging."""

from contextvars import ContextVar
from functools import wraps
from logging import LoggerAdapter, getLogger as get_logger
from typing import Any, Callable, List, MutableMapping, Optional, Tuple

from frgal.utils import get_call_id

# List of (random) strings to create log message prefix for nested contexts
# Note: It should only be manipulated with using dedicated methods in logger adapter.
log_ctx: ContextVar[List[str]] = ContextVar("log_ctx", default=[])


class LoggerContextError(Exception):
    """Invalid call to leave logger context."""


class PainLogger(LoggerAdapter):
    """Customized logger class."""

    @staticmethod
    def enter_context(name: str) -> None:
        """Enter new context."""
        ctx = log_ctx.get()
        ctx.append(name)
        log_ctx.set(ctx)

    @staticmethod
    def leave_context() -> None:
        """Leave current context and return to previous one."""
        ctx = log_ctx.get()
        if not ctx:
            raise LoggerContextError("No previous logger context available")
        ctx.pop()
        log_ctx.set(ctx)

    def process(self, msg: str, kwargs: MutableMapping[str, Any]) -> Tuple[str, MutableMapping[str, Any]]:
        """Process logging message."""
        msg, kwargs = super().process(msg, kwargs)
        ctx = ":".join(log_ctx.get())
        if ctx:
            msg = f"[{ctx}] {msg}"
        return msg, kwargs


def getLogger(name: Optional[str]) -> PainLogger:
    """Get customized logger instance."""
    return PainLogger(get_logger(name), {})


class NestedLoggingContext:
    """Context manager to do logging in new nested context."""

    def __enter__(self):
        PainLogger.enter_context(get_call_id())

    def __exit__(self, exc_type, exc_val, exc_tb):
        PainLogger.leave_context()


def nested_logging_context(func: Callable) -> Callable:
    """Decorator to do logging in new nested context."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        with NestedLoggingContext():
            return func(*args, **kwargs)

    return wrapper
