#
# Copyright (C) 2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Tests of customized logging."""

from unittest.mock import Mock, patch

from django.test import SimpleTestCase
from testfixtures import LogCapture

from django_pain.logging import (
    LoggerContextError,
    NestedLoggingContext,
    PainLogger,
    getLogger,
    log_ctx,
    nested_logging_context,
)


class TestPainLogger(SimpleTestCase):
    """Tests of logging adapter."""

    @staticmethod
    def do_some_logging(logger: PainLogger) -> None:
        logger.debug("debug")
        logger.info("info")
        logger.warning("warning")
        logger.error("error")
        logger.critical("critical")

    def setUp(self) -> None:
        self.log_handler = LogCapture("test_logging", propagate=False)
        log_ctx.set([])

    def tearDown(self) -> None:
        self.log_handler.uninstall()
        log_ctx.set([])

    def test_logging_empty_context(self):
        self.do_some_logging(getLogger("test_logging"))

        self.log_handler.check(
            ("test_logging", "DEBUG", "debug"),
            ("test_logging", "INFO", "info"),
            ("test_logging", "WARNING", "warning"),
            ("test_logging", "ERROR", "error"),
            ("test_logging", "CRITICAL", "critical"),
        )

    def test_logging_with_context(self):
        logger = getLogger("test_logging")
        logger.enter_context("abcd")

        self.do_some_logging(logger)

        logger.leave_context()
        self.log_handler.check(
            ("test_logging", "DEBUG", "[abcd] debug"),
            ("test_logging", "INFO", "[abcd] info"),
            ("test_logging", "WARNING", "[abcd] warning"),
            ("test_logging", "ERROR", "[abcd] error"),
            ("test_logging", "CRITICAL", "[abcd] critical"),
        )

    def test_logging_with_changing_context(self):
        logger = getLogger("test_logging")

        logger.info("Event 1")
        logger.enter_context("0123")
        logger.info("Event 2")
        logger.enter_context("4567")
        logger.info("Event 3")
        logger.enter_context("89ab")
        logger.info("Event 4")
        logger.leave_context()
        logger.info("Event 5")
        logger.leave_context()
        logger.info("Event 6")
        logger.leave_context()
        logger.info("Event 7")

        self.log_handler.check(
            ("test_logging", "INFO", "Event 1"),
            ("test_logging", "INFO", "[0123] Event 2"),
            ("test_logging", "INFO", "[0123:4567] Event 3"),
            ("test_logging", "INFO", "[0123:4567:89ab] Event 4"),
            ("test_logging", "INFO", "[0123:4567] Event 5"),
            ("test_logging", "INFO", "[0123] Event 6"),
            ("test_logging", "INFO", "Event 7"),
        )

    def test_logging_context_error(self):
        logger = getLogger("test_logging")

        self.assertRaisesMessage(LoggerContextError, "No previous logger context available", logger.leave_context)

    @patch("django_pain.logging.get_call_id", Mock(return_value="abcd"))
    def test_logging_with_context_manager(self):
        logger = getLogger("test_logging")

        with NestedLoggingContext():
            self.do_some_logging(logger)

        self.assertEqual(log_ctx.get(), [])
        self.log_handler.check(
            ("test_logging", "DEBUG", "[abcd] debug"),
            ("test_logging", "INFO", "[abcd] info"),
            ("test_logging", "WARNING", "[abcd] warning"),
            ("test_logging", "ERROR", "[abcd] error"),
            ("test_logging", "CRITICAL", "[abcd] critical"),
        )

    @patch("django_pain.logging.get_call_id", Mock(return_value="abcd"))
    def test_context_manager_exception(self):
        with self.assertRaisesMessage(Exception, "Heeeelp!"):
            with NestedLoggingContext():
                self.assertEqual(log_ctx.get(), ["abcd"])
                raise Exception("Heeeelp!")

        self.assertEqual(log_ctx.get(), [])

    @patch("django_pain.logging.get_call_id", Mock(side_effect=["0123", "4567"]))
    def test_logging_with_multiple_context_managers(self):
        logger = getLogger("test_logging")

        with NestedLoggingContext():
            logger.info("Event 1")
            with NestedLoggingContext():
                logger.info("Event 2")

        self.assertEqual(log_ctx.get(), [])
        self.log_handler.check(
            ("test_logging", "INFO", "[0123] Event 1"),
            ("test_logging", "INFO", "[0123:4567] Event 2"),
        )

    @patch("django_pain.logging.get_call_id", Mock(return_value="0123"))
    def test_logging_with_decorator(self):
        @nested_logging_context
        def log_with_decorator(logger: PainLogger) -> None:
            logger.info("Event")

        logger = getLogger("test_logging")

        log_with_decorator(logger)

        self.assertEqual(log_ctx.get(), [])
        self.log_handler.check(("test_logging", "INFO", "[0123] Event"))

    @patch("django_pain.logging.get_call_id", Mock(return_value="0123"))
    def test_logging_with_decorator_exception(self):
        @nested_logging_context
        def log_with_decorator(logger: PainLogger) -> None:
            logger.info("Event")
            raise Exception("Heeeelp!")

        logger = getLogger("test_logging")

        with self.assertRaisesMessage(Exception, "Heeeelp!"):
            log_with_decorator(logger)

        self.assertEqual(log_ctx.get(), [])
        self.log_handler.check(("test_logging", "INFO", "[0123] Event"))
